/* written by John Doug Reynolds, March 2002 */

#include "rgsenergy.h"
#include <PixelsTable.h>
#include <CalServer.h>
#include <Dal.h>
#include "OalServer.h"
#include "Xmm.h"

void rgsenergy()
{
   string dsname = stringParameter("ccdset");
   bool withoffset = booleanParameter("withoffset");
   bool withdiagoffset = booleanParameter("withdiagoffset");
   bool withgain = booleanParameter("withgain");
   bool withcti = booleanParameter("withcti");

   Xmm::Node cnode = static_cast<Xmm::Node>(0);
   Xmm::Node dnode = static_cast<Xmm::Node>(1);

   if ( withoffset ) {
      string offset = stringParameter("offset");
      calServer.ccf()->registerNewCifEntry( offset );
   }

   struct DataSet *dataset = dataSetServer->open( dsname, Dal::Modify );
   calServer << *dataset;
   unsigned ccdid = calServer.state()->ccd()->get();

   string sinst(dataset->attribute("INSTRUME")->asString());
   Xmm::Instrument instrument = 
         Xmm::toInstrument(dataset->attribute("INSTRUME")->asString());
    
   struct EnergyCorrector* calEC = 0; calEC = calServer.getAtom(calEC);
   struct CtiCorrector *calCC = 0;
   if (withcti)
      calCC= calServer.getAtom(calCC);

   {
      using namespace rgsPixelsLib;

      struct Table *ptab = dataset->table("PIXELS");
      struct Table *ftab = dataset->table("EXPOSURE");

      if ( !ptab->hasColumn("PI") )
	 ptab->addColumn( "PI", Dal::Real32, "calibrated energy" );

      unsigned flags = PixelsTable::PIX | PixelsTable::PHA | PixelsTable::CPI;
      const PixelsTable pix(ptab,0,flags);

      const DalVecJ frame( ftab->column("FRAME") );
      const DalVecD  time( ftab->column("TIME") );

      // 

      CalReal32Vector calpiV(  pix.calpi.data(), pix.rows );
      CalInt16Vector   rawxV(   pix.rawX.data(), pix.rows );
      CalInt16Vector energyV( pix.energy.data(), pix.rows );
      CalInt8Vector    nodeV(   pix.node.data(), pix.rows );

      CalInt16Vector  rawyV(  pix.rawY.data(), pix.modeIsHTR()? 0 : pix.rows );
      CalInt8Vector  shapeV( pix.shape.data(), pix.modeIsHTR()? 0 : pix.rows );

      CalReal32Vector  tmpV( 0, pix.rows );
      CalReal64Vector timeV( 0, pix.rows );

      unsigned fItr = 0, pItr = 0;
      for (; fItr < ftab->rows() && pItr < pix.rows; ++fItr )
	 if ( frame[fItr] >= pix.frame[pItr] ) {
	    int sameFrame = pix.frame[pItr];
	    double t = time[ frame[fItr] == sameFrame ? fItr : 0 ];
	    while ( pItr < pix.rows && pix.frame[pItr] == sameFrame )
	       timeV[pItr++] = t;
	 }

      while ( pItr < pix.rows ) timeV[pItr++] = time[0];

      // 

			       
      unsigned i, c = 0;
      bool no_rgs_diag_map = false;

     static unsigned short trans[] = { 1, 4, 2, 2, 3, 3, 3, 3 }; 
	SImage diagnosticMapC;
	SImage diagnosticMapD;

      if ( pix.modeIsHTR() )
	 calEC->offsetCorrect(timeV,rawxV,nodeV,energyV,calpiV);

// CG pixel dependant offset ...........................

      else {
	 if ( withdiagoffset ) {

        	oalServer.instrument(instrument);
        	oalServer.ccd(ccdid);

        	oalServer.node(cnode);
	
		no_rgs_diag_map = false;
		if  (oalServer.odf()->hasSet(Xmm::SYSPEAK_DIAGNOSTIC))
		{
		   diagnosticMapC = oalServer.odf()->rgsDiagnosticMap();
        	   oalServer.node(dnode);
		   diagnosticMapD = oalServer.odf()->rgsDiagnosticMap();
		}
		else 
		{
		  no_rgs_diag_map = true;    
		  errstr << "No Diagnostic Offset Average File found for " << sinst << endl
		         << "I am going to use default value subtraction for it "
			 << Sas::message(Msg::App,Msg::Verbose);
		}
	}
	
	if ( withdiagoffset && !no_rgs_diag_map ) {	
		unsigned long yExtent = diagnosticMapC.yExtent();

        	for ( i = 0; i < pix.rows; ++i ) {
   	
	 	if ( nodeV[i] == 0 ) 
		       calpiV[i] = energyV[i] - ( trans[shapeV[i]] * 
	       		diagnosticMapC[rawyV[i]-1][rawxV[i]-1]);
		else
		       calpiV[i] = energyV[i] - ( trans[shapeV[i]] *
	       		diagnosticMapD[rawyV[i]-1][yExtent - rawxV[i]]);
	    	}

        }
//
	else
	
// Single value CAL offset correction 

	   calEC->offsetCorrect(timeV,rawxV,rawyV,shapeV,nodeV,energyV,calpiV);
     
      }

      CalReal32Vector* v[2] = { &calpiV, &tmpV };


      if ( withgain ) {
	 calEC->gainCorrect(timeV,nodeV,*v[c],*v[1-c]);
	 c = 1 - c;
      }

      if ( withcti ) {
	 if ( pix.modeIsHTR() )
	    calCC->ctiCorrect(timeV,rawxV,pix.htrY0,*v[c],*v[1-c],&nodeV);
	 else {
	    CalInt16Vector sepV( 0, pix.rows );
	    for ( i = 0; i < pix.rows; ++i ) sepV[i] = 1000; // dummy value
	    calCC->ctiCorrect(timeV,rawxV,rawyV,shapeV,sepV,
			      *v[c],*v[1-c],&nodeV);
	 }
	 c = 1 - c;
      }

      if ( c ) for ( i = 0; i < pix.rows; ++i ) pix.calpi[i] = tmpV[i];
   }

   dataSetServer->close( dataset );

   if ( withcti ) calServer.releaseAtom(calCC);
   calServer.releaseAtom(calEC);
}
